#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['flight_arbitrage']

package_data = \
{'': ['*']}

setup(name='flight_arbitrage',
      version='0.7.1',
      description='Finds arbitrage in plane ticket prices',
      author='Jordan Rinder',
      author_email='jordan.rinder@gmail.com',
      url='https://github.com/jrinder42/Flight_Arbitrage',
      packages=packages,
      package_data=package_data,
      python_requires='>=3.7',
     )
